// CRDataServer.cpp

#include "CRDataServer.hpp"
#include "utils/c++/SocketUtils.hpp"
#include "utils/c++/FileUtils.hpp"
// gethostbyname

#include <fcntl.h>

#include <iostream>
#include <fstream>
#include <sstream>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <strings.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

struct DataInfo {
  char *data;  
  long dataSize;
  char *execId;
  long execIdSize;  
  int ckpNumber;
  int fragmentNumber;
};

//------------------------------------------------------------------------------
CRDataServer::CRDataServer(AdrManagerStub & stub, const int & port) : adrManager(stub) {
    
    this->port = port;
    this->maxPort = this->port + 100;
  
    // sets and creates output dir
    this->outputDir = "/tmp/ckpRep/";
    mkdir(outputDir.c_str(), 0700);
        
    // obtains the ipAddress of the host
    system("hostname -i > machineIp.dat");
    ifstream ipFile("machineIp.dat");
    ipFile >> this->ipAddress;
    ipFile.close();
      
    // Initializes Server  
    this->setupServer();      
}

//------------------------------------------------------------------------------
CRDataServer::CRDataServer(AdrManagerStub & stub, string ipAddress, const int & port) : adrManager(stub) {

    this->port = port;
    this->maxPort = this->port + 100;
  
    // sets and creates output dir
    this->outputDir = "/tmp/ckpRep/";
    mkdir(outputDir.c_str(), 0700);
        
    // obtains the ipAddress of the host
    this->ipAddress = ipAddress;
      
    // Initializes Server  
    this->setupServer();      
}

//------------------------------------------------------------------------------
void CRDataServer::readStream(int sockfd) {

  long methodType = SocketUtils::readUint32(sockfd);

  // Store checkpoint data --> fragmentIdSize(int) fragmentId(char*) fragmentSize(int) fragmentData(char*)
  if (methodType == 1) { 

    std::ostringstream tempFileName;
    tempFileName << outputDir << getpid() << ".adr";
    int tempfd = open(tempFileName.str().c_str(), O_WRONLY | O_APPEND | O_CREAT, 0600);    
    
    // Reads the file data from the stream to 'data'
    int fragmentDataSize = (long)SocketUtils::readUint32(sockfd);
    int nWrittenBytes = SocketUtils::readFromStreamToFile(sockfd, fragmentDataSize, tempfd);
    
    close(tempfd);

    // Reads the fragmentId from the stream to 'fragmentId'
    int fragmentIdSize = (long)SocketUtils::readUint32(sockfd);        
    char *fragmentId = new char[fragmentIdSize+1];
    fragmentId[fragmentIdSize]=0; // Marks the end of the string
    SocketUtils::readFromStream(sockfd, fragmentId, fragmentIdSize, NULL);

    // Saves data to filesystem -----------------------
    std::ostringstream ckpFileName;
    ckpFileName << outputDir << fragmentId << ".adr";
    rename(tempFileName.str().c_str(), ckpFileName.str().c_str());
    // ------------------------------------------------
    
    int writeCompleted = 1;
    SocketUtils::writeUint32(sockfd, writeCompleted);

    std::cout << "Data written to " << ckpFileName.str() << " nbytes=" << nWrittenBytes << " of " << fragmentDataSize << "." << std::endl;

    // Notifies the ckpReposManager about the newly stored checkpoint
    int fifofd = open( fifoName_.c_str(), O_WRONLY );
    write( fifofd, &nWrittenBytes, sizeof(ssize_t) );

    //std::cout << "FileManipulationNotifier notified " << ckpFileName.str() << " nbytes=" << nbytes << " fifofd=" << fifofd << "." << std::endl;
       
    delete[] fragmentId;
    
    //std::cout << "Saved " << fragmentDataSize << " bytes to " << ckpFileName.str() << std::endl;                 
  }
  
  // Recover checkpointing data --> fragmentIdSize(int) fragmentId(char*) | fragmentSize(int) fragmentData(char*)
  else if (methodType == 2) {

    // Reads the executionId from the stream to 'data'
    long fragmentIdSize = (long)SocketUtils::readUint32(sockfd);        
    char *fragmentId = new char[fragmentIdSize+1];
    fragmentId[fragmentIdSize]=0;      
    SocketUtils::readFromStream(sockfd, fragmentId, fragmentIdSize, NULL);
    
    // Open the correct file and put in buffer --------
    std::ostringstream ckpFileName;
    ckpFileName << outputDir << fragmentId << ".adr";
    int inFile = open(ckpFileName.str().c_str(), O_RDONLY);
    if (inFile == -1) 
      std::cerr << "Couldn't open file " << ckpFileName.str() << "." << std::endl;
    struct stat *fileStat = new struct stat;
    fstat(inFile, fileStat);
    long fragmentDataSize = (long)fileStat->st_size;
    delete fileStat;
    char *fragmentData = new char[fragmentDataSize];
    read(inFile, fragmentData, fragmentDataSize);
    // -------------------------------------------------
    
    std::cout << "Returning Data from " << ckpFileName.str() << " dataSize=" << fragmentDataSize << std::endl;
        
    // Writes the number of bytes of the file
    SocketUtils::writeUint32(sockfd, (long)fragmentDataSize);
    // Transfer data to repository
    SocketUtils::writeToStream(sockfd, fragmentData, fragmentDataSize);
    
    delete[] fragmentData;
    delete[] fragmentId;
  }
  
}

//------------------------------------------------------------------------------

void CRDataServer::setupServer() {
  
  // Open a TCP socket
  this->sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if ( sockfd < 0 )
    std::cerr << "CRDataServer::startServer --> cannot open stream socket." << std::endl;

  // Binds the server address
  struct sockaddr_in serverAddress;
  bzero((char *) &serverAddress, sizeof(serverAddress));
  serverAddress.sin_family      = AF_INET;
  serverAddress.sin_addr.s_addr = htonl(INADDR_ANY);  
  serverAddress.sin_port        = htons(port);
  
  bool bound = false;
  while (bound == false && port < maxPort) {  
    if ( bind(sockfd, (struct sockaddr *)&serverAddress, sizeof(serverAddress)) == 0 )
      bound = true;
    else
      serverAddress.sin_port = htons(++port);
  }
            
  ostringstream oppStoreAddress;
  oppStoreAddress << this->ipAddress << ":" << this->port;
  int availableSpace = FileUtils::getAvailableDiskSpace(this->outputDir.c_str());
  this->ckpReposId = adrManager.registerAdr(oppStoreAddress.str(), availableSpace, 1.0, 1.0);
  
  ostringstream fifoNameStream;
  fifoNameStream << this->outputDir << "/adrWrite.fifo";
  this->fifoName_ = fifoNameStream.str(); 
  mkfifo( this->fifoName_.c_str(), S_IRUSR | S_IWUSR );
  //std::cerr << "ERROR! Could not create FIFO file " << this->fifoName_ << "." << std::endl;
  
  std::cout << "ckpRepository: Listening on ip:" << ipAddress      << " port:" << port << std::endl;      
  std::cout << "ckpRepository: started with id " << ckpReposId     << "."      << std::endl;  
  std::cout << "ckpRepository: saving data at '" << outputDir      << "'."     << std::endl;
  std::cout << "ckpRepository: availableSpace="  << availableSpace << "."      << std::endl;
   
  listen(sockfd, 128);          
}

//----------------------------------------------------------------------------

void CRDataServer::startServer() {
    
  while(true) {
    
    // Accepts the connection from a client
    struct sockaddr_in clientAddress;
    socklen_t clientAddressLength = sizeof(clientAddress);
    int newsockfd = 
      accept (sockfd, (struct sockaddr *)&clientAddress, &clientAddressLength);
    
    if (newsockfd < 0)
      std::cerr << "CRDataServer::startServer --> error accepting connection." << std::endl;
    
    // Forks a new process for concurrent processing
    int pidMid = fork();
    if (pidMid == 0) {  //mid
      int pidBotton = fork();
      if (pidBotton == 0) {  //son
        close (sockfd); // close original socket
        this->readStream(newsockfd);
        exit(0);
      }
      else 
	    exit(0);
    }
    else
        wait(NULL);

    // Parent process: close the new socket
    close (newsockfd); 
  }
}

//------------------------------------------------------------------------------
