#include "AdrDataTransferStub.hpp"
#include "utils/c++/SocketUtils.hpp"
#include "utils/c++/StringUtils.hpp"

#include <unistd.h>
#include <fcntl.h>
#include <cassert>

using namespace std;

//===========================================================================
// CkpRepositoryStub
//===========================================================================

int AdrDataTransferStub::connectToServer(string ipAddress, short port) {

  // Fills serverAddress with the server address and port
  struct sockaddr_in serverAddress;
  bzero((char *) &serverAddress, sizeof(serverAddress));
  serverAddress.sin_family      = AF_INET;
  serverAddress.sin_addr.s_addr = inet_addr(ipAddress.c_str());  
  serverAddress.sin_port        = htons(port);
 
  // Opens a TCP socket
  int sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if ( sockfd < 0 ) {
    std::cerr << "CkpRepositoryStore::connectToServer --> cannot open stream socket." << std::endl;
    return -1;
  }

  // Connects to server
  int connectFlag = connect (sockfd, (struct sockaddr *)&serverAddress, sizeof(serverAddress));
  if ( connectFlag < 0 ) {
    std::cerr << "CkpRepositoryStore::connectToServer --> cannot connect to server." << std::endl;
    return -1;
  }
   
  return sockfd;
}

//---------------------------------------------------------------------------
  
int AdrDataTransferStub::prepareDataTransfer(const string & ipAddress, short portNumber) {

    int sockfd = this->connectToServer(ipAddress, portNumber);  
    if (sockfd < 0) return -1;
  
    // sending a new fragment
    SocketUtils::writeUint32(sockfd, (long)1); // Indicates that a store will be performed
  
    return sockfd;
}

int AdrDataTransferStub::performDataTransfer(int sockfd, const void *data, long nbytes) {

    // Transfer data to repository
    SocketUtils::writeUint32(sockfd, (long)nbytes);
    SocketUtils::writeToStream(sockfd, data, nbytes);
    
    return 0;
}

int AdrDataTransferStub::performDataTransferInc(int sockfd, const void *data, long nbytes, long *availableBytes) {

    // Transfer data to repository
    SocketUtils::writeUint32(sockfd, (long)nbytes);
    SocketUtils::writeToStreamInc(sockfd, data, nbytes, availableBytes);
    
    return 0;
}

int AdrDataTransferStub::finishDataTransfer(int sockfd, const char *fragmentId, int fragmentIdSize) {

  // Transfer the fragmentId
  SocketUtils::writeUint32(sockfd, (long)fragmentIdSize);
  SocketUtils::writeToStream(sockfd, fragmentId, fragmentIdSize);
  
  // get the number of bytes written
  long writeCompleted = SocketUtils::readUint32(sockfd);
  
  close (sockfd);

  return writeCompleted;
}

//---------------------------------------------------------------------------
  
int AdrDataTransferStub::transferData
(const string & ipAddress, short portNumber, const char *fragmentId, int fragmentIdSize, const void *data, long nbytes) {

    int sockfd = prepareDataTransfer(ipAddress, portNumber);
    performDataTransfer(sockfd, data, nbytes);
    return finishDataTransfer(sockfd, fragmentId, fragmentIdSize);
    
}

//---------------------------------------------------------------------------

int AdrDataTransferStub::readData(const string & ipAddress, short portNumber, const char *fragmentId, int fragmentIdSize, void * & data, int & dataSize, long *availableBytes) {

  int sockfd = this->connectToServer(ipAddress, portNumber);
  if (sockfd < 0) return -1;
    
  // Asking for checkpoint 'CkpNumber' 
  SocketUtils::writeUint32(sockfd, (long)2);
  
  // Transfer the executionId
  SocketUtils::writeUint32(sockfd, (long)fragmentIdSize);
  SocketUtils::writeToStream(sockfd, fragmentId, fragmentIdSize);

  // Reads the file data from the stream to 'data'
  long tmpSize = SocketUtils::readUint32(sockfd);
  if (dataSize < 0) {
    dataSize = tmpSize;      
    data = malloc(dataSize);
  }

  SocketUtils::readFromStream(sockfd, data, (long)dataSize, availableBytes);
  close (sockfd);  
  std::cerr << "AdrDataTransferStub::readData -> Finished recovering data!!! nbytes=" << dataSize << std::endl;
  
  return tmpSize;
}

//---------------------------------------------------------------------------

