#include "DataStorageManager.hpp"
#include "FragmentUploadThread.hpp"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"

DataStorageManager::DataStorageManager()
{
}

DataStorageManager::~DataStorageManager()
{
}

void DataStorageManager::setCdrmRequestStub( CdrmRequestsStub *cdrmRequestsStub, AdrDataTransferStub *adrDataTransferStub ) {
 
    this->cdrmRequestsStub_    = cdrmRequestsStub;
    this->brokerIor_           = AccessBrokerSkeleton::singleInstance().getIor();  
    this->adrDataTransferStub_ = adrDataTransferStub;
    
    FragmentUploadThread::configureUploadThreads(cdrmRequestsStub, adrDataTransferStub, this);
}
       
void DataStorageManager::storeData( CodedDataInfo *dataInfo ) {
    
    int timeoutMinutes = 1;
    bool storeGlobal = true;
    int requestId = cdrmRequestsStub_->requestFileStorage(
                        dataInfo->fileKey, dataInfo->fragmentKeyList, dataInfo->dataSize, dataInfo->fragmentSizeList,
                        this->brokerIor_, timeoutMinutes, storeGlobal);                                     
    requestInfoMap[requestId] = dataInfo;

    cout << "Requested data storage with requestId=" << requestId << "." << endl;
}
    
CodedDataInfo *DataStorageManager::getCodedDataInfo ( int requestId ) {
    
    CodedDataInfo *dataInfo = NULL;
    for (int i=0; i<10 && dataInfo == NULL; i++) {
        if ( requestInfoMap.find(requestId) != requestInfoMap.end() )
            dataInfo = requestInfoMap[requestId];
        else
            OppStoreUtils::sleep(100*(i+1));
    }               
    return dataInfo;
}
        
void DataStorageManager::uploadFragments ( int requestId, vector<string> adrAddresses ) {
        
    cout << "Received upload fragment message! requestId=" << requestId << "." << endl;

    CodedDataInfo *dataInfo = this->getCodedDataInfo( requestId ); 
    dataInfo->adrAddresses.assign( adrAddresses.begin(), adrAddresses.end() );
                        
//    for (unsigned int i=0; i<adrAddresses.size(); i++)
//        cout << adrAddresses[i] << " ";
//    cout << endl;
   
    /**
     * Waits for the coding process to finish
     */   
    //OppStoreUtils::waitVar(100, dataInfo->remainingCodings);
    //Benchmark::getCurrentBenchmark()->requestFree = OppStoreUtils::timeInMillis();
    
    /**
     * Uploads fragment data, launching one thread per fragment.
     */
    int * requestIdPtr = new int;
    *requestIdPtr = requestId;
    //FragmentUploadThread::launchUploadThread(requestIdPtr);
    pthread_t thread1;
    pthread_create( &thread1, NULL, FragmentUploadThread::launchUploadThread, requestIdPtr );
    pthread_detach(thread1);

}    
        
void DataStorageManager::setFileStorageRequestCompleted( int requestId ) {

    cout << "File storage completed for request " << requestId << "." << endl;

    CodedDataInfo *dataInfo = requestInfoMap.find(requestId)->second;
    if (dataInfo->operationCallback != NULL)
        dataInfo->operationCallback->finishedDataOperation(dataInfo, 0);

}
