#include "FragmentUploadThread.hpp"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"

#include <openssl/sha.h>

#include <map>
#include <iostream>
#include <cassert>
using namespace std;

map<int, FragmentUploadThread *> uploadThreadMap;
int nextUploadThreadId = 1;

CdrmRequestsStub *cdrmRequestsStub;
AdrDataTransferStub *adrDataTransferStubUpload;
DataStorageManager *dataStorageManager;

void FragmentUploadThread::configureUploadThreads( CdrmRequestsStub *cdrmRequestsStub_, AdrDataTransferStub *adrDataTransferStub_, DataStorageManager *dataStorageManager_ ) {

    cdrmRequestsStub = cdrmRequestsStub_;
    adrDataTransferStubUpload = adrDataTransferStub_;
    dataStorageManager = dataStorageManager_;

}


FragmentUploadThread::FragmentUploadThread
( string ipAddress_, short portNumber_, unsigned char *fragmentKey_, char *encData_, int fragmentSize_, int fragmentNumber_, int *remainingUploads_, long *availableBytes_)
{

    this->ipAddress = ipAddress_;
    this->portNumber = portNumber_;
    
    this->fragmentKey = fragmentKey_;
    
    this->encData = encData_;
    this->fragmentSize = fragmentSize_;    
    this->fragmentNumber = fragmentNumber_;
    
    this->availableBytes = availableBytes_;
    
    this->remainingUploads = remainingUploads_;
}

FragmentUploadThread::~FragmentUploadThread()
{
}

void FragmentUploadThread::performUpload() {

    Benchmark::getCurrentBenchmark()->fragmentStart[fragmentNumber] = OppStoreUtils::timeInMillis();
    
    cout << "DataStorageManager -> uploading fragment to " << ipAddress << ":" << portNumber << " availableBytes=" << *availableBytes << "." << endl;

    //adrDataTransferStub->transferDataInc(ipAddress, portNumber, hexKey, OppStoreUtils::binaryKeySize*2, encData, fragmentSize, availableBytes);

    int sockfd = adrDataTransferStubUpload->prepareDataTransfer(ipAddress, portNumber);
    adrDataTransferStubUpload->performDataTransferInc(sockfd, encData, fragmentSize, availableBytes);
    
    SHA1( (unsigned char *)encData, fragmentSize, fragmentKey);                   
    char *hexKey = new char[ OppStoreUtils::binaryKeySize*2 ];
    OppStoreUtils::convertBinaryToHex((char *)fragmentKey, hexKey);

    adrDataTransferStubUpload->finishDataTransfer(sockfd, hexKey, OppStoreUtils::binaryKeySize*2);

    delete[] hexKey;
    
    (*remainingUploads)--;
    cout << "remainingUploads=" << *remainingUploads << endl;

    Benchmark::getCurrentBenchmark()->fragmentFinish[fragmentNumber] = OppStoreUtils::timeInMillis();
}

void *FragmentUploadThread::run( void *ptr ) {
 
    int *threadId = (int *)ptr;
    FragmentUploadThread *uploadThread = uploadThreadMap[*threadId];
    uploadThread->performUpload();
    
    uploadThreadMap.erase(*threadId);
    delete threadId;

    return NULL;
}

void *FragmentUploadThread::launchUploadThread ( void *ptr ) {

    int requestId = *(int *)ptr;
    delete (int *)ptr;
    CodedDataInfo *dataInfo = dataStorageManager->getCodedDataInfo(requestId );
    
    /**
     * TODO: Deal with the case that some repositories do not work.
     */    
    vector<int> notStoredFragments;
    vector<char *> fragmentHashList;

    int nAddresses = dataInfo->adrAddresses.size();
    dataInfo->remainingUploads = new int;
    *(dataInfo->remainingUploads) = nAddresses;
    
    Benchmark *benchmark = Benchmark::getCurrentBenchmark();
    benchmark->nFragmentTimes = nAddresses;
    benchmark->fragmentStart  = new long[nAddresses];
    benchmark->fragmentFinish = new long[nAddresses];
    benchmark->fragmentAddress.resize(nAddresses);

    for (int fragment=0; fragment < nAddresses; fragment++) {

        string ipAddress;
        short portNumber;
        OppStoreUtils::extractIpAddress(dataInfo->adrAddresses[fragment], ipAddress, portNumber);

        unsigned char *fragmentKey = new unsigned char[OppStoreUtils::binaryKeySize]; 
        fragmentHashList.push_back( (char *)fragmentKey );
    
        Benchmark::getCurrentBenchmark()->fragmentAddress[fragment] = dataInfo->adrAddresses[fragment];
                
        /**
         * Launches the fragment upload thread.
         */ 
        FragmentUploadThread *uploadThread = 
            new FragmentUploadThread( ipAddress, portNumber, fragmentKey, dataInfo->encData[fragment], dataInfo->fragmentSizeList[fragment], 
                                      fragment, dataInfo->remainingUploads, dataInfo->availableBytes[fragment] );
        int *threadId = new int;
        *threadId = nextUploadThreadId++;
        uploadThreadMap[*threadId] = uploadThread;

        pthread_t thread1;
        pthread_create( &thread1, NULL, FragmentUploadThread::run, threadId );
        pthread_detach(thread1);
    }
    
    OppStoreUtils::waitVar(100, dataInfo->remainingUploads);
                        
    cdrmRequestsStub->setFragmentStorageFinished(requestId, notStoredFragments, fragmentHashList);
    
    return NULL;
}

