/**
 * Performs the storage and recovery of a file and tests if the file was recovered correctly
 */
  
#include "OppStoreBroker.h"
#include "OppStoreUtils.hpp"
#include "Benchmark.hpp"

#include <cstdlib>
#include <cassert>
#include <string>
#include <iostream>

#include <openssl/sha.h>

void testConversion() {
    long dataSize = 100;
    char *data = new char [dataSize];
    for (int i=0; i<dataSize; i++)
        data[i] = rand()%256;            
    
    unsigned char *fileKey1   = new unsigned char[OppStoreUtils::binaryKeySize];
    unsigned char *fileKeyHex = new unsigned char[OppStoreUtils::binaryKeySize*2];
    unsigned char *fileKey2   = new unsigned char[OppStoreUtils::binaryKeySize];    
    SHA1( (unsigned char*)data, dataSize, (unsigned char*)fileKey1 );
    OppStoreUtils::convertBinaryToHex((char *)fileKey1,   (char *)fileKeyHex);
    OppStoreUtils::convertHexToBinary((char *)fileKeyHex, (char *)fileKey2);    
    for (int i=0; i<OppStoreUtils::binaryKeySize; i++)
        assert( fileKey1[i] == fileKey2[i] );
}

char *testFileStorage(char *data, long dataSize) {

    std::cout << "testFileStorage -> Started storing data with size " << dataSize << "!" << std::endl;        
    char *key;
    storeDataW(key, data, dataSize, NULL);
    std::cout << "testFileStorage -> Finished storing data!" << std::endl;
    
    return key;
}

void testFileRetrieval(char *data, long dataSize, char *key) {
    
    std::cout << "testFileRetrieval -> Started recovering data!" << std::endl;
    void *recoveredData;
    long recoveredSize;    
    retrieveDataW( key, recoveredData, recoveredSize, NULL );
    std::cout << "testFileRetrieval -> Finished recovering data! nbytes=" << recoveredSize << std::endl;
    
    unsigned char *fileKey1 = new unsigned char[OppStoreUtils::binaryKeySize];
    unsigned char *fileKey2 = new unsigned char[OppStoreUtils::binaryKeySize];    
    SHA1(          (unsigned char*)data, dataSize, (unsigned char*)fileKey1);
    SHA1( (unsigned char*)recoveredData, dataSize, (unsigned char*)fileKey2);

    for (int i=0; i<dataSize; i++)
        assert( data[i] == ((char *)recoveredData)[i] );
    
    for (int i=0; i<OppStoreUtils::binaryKeySize; i++)
        assert( fileKey1[i] == fileKey2[i] );
     
     std::cout << "Removing recovered data!" << std::endl;   
     delete[] (char *)recoveredData;
     delete[] fileKey1;
     delete[] fileKey2;     
}

int main(int argc, char **argv) {

    launchBroker();
    std::cout << "Finished launching broker!" << std::endl;
    
    srand(12345);
    
    testConversion();
    
    for (int i=0; i<1; i++) {
        //long dataSize = rand()%1000;
        long dataSize = 3000;
        char *data = new char [dataSize];
        for (int i=0; i<dataSize; i++)
            data[i] = rand()%256;

        /**
         * Tests file storage
         */            
        Benchmark *benchmark = Benchmark::createNewBenchmark();
        benchmark->benchmarkId = 1;                        
        benchmark->dataSize = dataSize; 
        benchmark->requestStart = OppStoreUtils::timeInMillis();
        
        char *key = testFileStorage(data, dataSize);
        
        benchmark->requestFinish = OppStoreUtils::timeInMillis();
        Benchmark::writeBenchmarkToFile(benchmark, "/tmp/bench.dat");

        /**
         * Tests file retrieval
         */
        benchmark = Benchmark::createNewBenchmark();
        benchmark->benchmarkId = 2;
        //benchmark->dataSize = dataSize;
        benchmark->requestStart = OppStoreUtils::timeInMillis();

        testFileRetrieval(data, dataSize, key);

        benchmark->requestFinish = OppStoreUtils::timeInMillis();
        Benchmark::writeBenchmarkToFile(benchmark, "/tmp/bench.dat");

        delete[] data;
    }

    std::cout << "Test finished succesfully!" << std::endl;

    sleep(2);

    return 0;
}
