package br.usp.ime.oppstore.cdrm;

import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import br.usp.ime.oppstore.FileFragmentIndex;
import br.usp.ime.oppstore.adrmanager.AdrInformationStructure;

import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.NodeHandleSet;

/**
 *  Perform management of data stored in the CDRM.
 *  It contains the locations of fragments and the FFI contents.
 */
public class CdrmDataManager {
       
    /**
     * Maintains the addresses of CDRMs storing replicas of FileInformationStructures
     * nodeId x cdrmHandlesSet 
     */
    HashMap <Id, NodeHandleSet> fileInformationLocationMap;

    /**
     * Maintains the FileInformationStructure for a given fileId 
     * nodeId x fileInfo
     */
    HashMap <Id, FileFragmentIndex> fileInformationStructureMap;
    
    public CdrmDataManager () {
        fileInformationLocationMap = new HashMap <Id, NodeHandleSet>();
        fileInformationStructureMap = new HashMap <Id, FileFragmentIndex>() ;        
    }
        
    public void setFileInformationLocations (Id fileId, NodeHandleSet cdrmHandleSet) {
        fileInformationLocationMap.put(fileId, cdrmHandleSet);
    }

    /**
     * Stores a FileInformationStructure
     * @param fileId The Id of the FileInformationStructure
     * @param fileInformationStructure The FileInformationStructure that will be stored
     */    
    public void storeFileInformationStructure (Id fileId, FileFragmentIndex fileInfo) {
        fileInformationStructureMap.put(fileId, fileInfo);
    }
    
    /**
     * Gets a FileInformationStructure
     * TODO: Return a value different from null when FileInformationStructure is not found
     * 
     * @param fileId The Id of the FileInformationStructure 
     * @return The FileInformationStructure corresponding to the Id 'messageId'
     */
    public FileFragmentIndex getFileFragmentIndex(Id fileId) {        
        return fileInformationStructureMap.get(fileId);
    }
    
    public List<FileFragmentIndex> getFileInformationStructureList() {
        List<FileFragmentIndex> fileInfoList = 
            new Vector<FileFragmentIndex>(fileInformationStructureMap.size());
        
        for (FileFragmentIndex fileInfo : fileInformationStructureMap.values())
            fileInfoList.add(fileInfo);

        return fileInfoList;                
    }
}
