package br.usp.ime.oppstore.cdrm;

import java.util.Vector;

import org.omg.CORBA.ORB;

import br.usp.ime.oppstore.corba.AccessBroker;
import br.usp.ime.oppstore.corba.AccessBrokerHelper;
import br.usp.ime.oppstore.corba.CdrmRequestsPOA;

import rice.environment.Environment;
import rice.p2p.commonapi.Id;
import rice.p2p.commonapi.IdFactory;
import rice.pastry.commonapi.PastryIdFactory;

public class CdrmRequestsImpl extends CdrmRequestsPOA {
    
	private ORB orb;
    private FileStorageRetrievalManager fileStorageManager;
    protected IdFactory idFactory;
       
    public CdrmRequestsImpl (FileStorageRetrievalManager fileStorageManager, Environment environment, ORB orb) {
        this.fileStorageManager = fileStorageManager;   
        this.idFactory = new PastryIdFactory(environment);
        this.orb = orb;
    }        
    
    /**
     * From CDRM IDL interface 
     * Used to notify the CDRM that the fragments from file 'fileId' have been stored.
     * 
     * @param fileKey A byte sequence representing the file key
     * @param fragmentKeyList List of keys from fragments that couldn't be stored. 
     */
    public void setFragmentStorageFinished(int requestNumber, int[] notStoredFragmentIndexList, byte[][] fragmentHashList) {
                               
    	fileStorageManager.storeFileInformationIndex(requestNumber, notStoredFragmentIndexList, fragmentHashList);
    }
    
    /**
     * From CDRM IDL interface 
     * 
     * @param fragmentKeyArray A list of byte sequences representing keys 
     */
    public int requestFileStorage(byte[] fileKey, byte[][] fragmentKeyArray, int fileSize, int[] fragmentSizeArray, 
    		String accessBrokerIor, int timeoutMinutes, boolean storeGlobal) {

    	AccessBroker source = AccessBrokerHelper.narrow( orb.string_to_object( accessBrokerIor ) );
    	
        // generates Ids from the byte sequences
        Id fileId = idFactory.buildId(fileKey);
        Vector<Id> fragmentIdList = new Vector<Id>();
        for (byte[] key : fragmentKeyArray)
            fragmentIdList.add(idFactory.buildId(key));        
        
        int requestId = fileStorageManager.getRequestNumber(source);
        fileStorageManager.requestFragmentStorageLocations(requestId, fileSize, fragmentSizeArray, fileId, fragmentIdList, timeoutMinutes, storeGlobal);
                
        return requestId;
    }
    
    /**
     * From CDRM IDL interface 
     * 
     * @param fragmentKeyList A list of byte sequences representing keys 
     */
    public int requestFileRetrieval(byte[] fileKey, String accessBrokerIor) {

    	AccessBroker source = AccessBrokerHelper.narrow( orb.string_to_object( accessBrokerIor ) );

        Id fileId = idFactory.buildId(fileKey);
        
        return fileStorageManager.retrieveFileInformationIndex(fileId, source);
    }

}
