package br.usp.ime.oppstore.cdrm;

import br.usp.ime.oppstore.message.FileFragmentIndexMessage;
import br.usp.ime.virtualId.VirtualNode;
import rice.p2p.commonapi.NodeHandle;
import rice.p2p.commonapi.NodeHandleSet;

public class FileIndexFragmentManager {

	private VirtualNode virtualNode;
    protected CdrmDataManager cdrmDataManager;
	
	public FileIndexFragmentManager( VirtualNode virtualNode ) {
		this.virtualNode = virtualNode;
        this.cdrmDataManager = new CdrmDataManager();
	}
	
    /**
     * @param fileFragmentIndexMessage
     */
    public void processFileIndexMessage(FileFragmentIndexMessage fileFragmentIndexMessage) {

    	/**
    	 * FileFragmentIndex storage request
    	 */
        if (fileFragmentIndexMessage.isStoreRequest == true) {

        	/**
        	 * Forwards replica storage request to the closest neighbor nodes
        	 */
        	if (fileFragmentIndexMessage.isReplicaStoreRequest == false) {   

        		NodeHandleSet neighborNodesSet = virtualNode.getNode().getLeafSet().neighborSet(fileFragmentIndexMessage.numberOfReplicas);
        		for (int handleIndex = 1; handleIndex < fileFragmentIndexMessage.numberOfReplicas; handleIndex++) {
        			FileFragmentIndexMessage replicaFFIMessage = new FileFragmentIndexMessage(fileFragmentIndexMessage);
        			replicaFFIMessage.isReplicaStoreRequest = true;
        			
        			NodeHandle targetHandle = neighborNodesSet.getHandle(handleIndex);
        			if (targetHandle != null)
        				virtualNode.sendDirectMessage( targetHandle, replicaFFIMessage );
        			else {
        	        	replicaFFIMessage.isResponse = true;
        	        	virtualNode.sendDirectMessage(replicaFFIMessage.sourceHandle, replicaFFIMessage);
        			}
        		}
        	}

        	/**
        	 * stores data on local repository
        	 */
        	fileFragmentIndexMessage.fileFragmentIndex.isReplica = fileFragmentIndexMessage.isReplicaStoreRequest;
        	cdrmDataManager.storeFileInformationStructure(fileFragmentIndexMessage.fileId, fileFragmentIndexMessage.fileFragmentIndex);

        	/**
        	 * returns FileFragmentIndex to sourceCdrm
        	 */
        	fileFragmentIndexMessage.isResponse = true;
        	virtualNode.sendDirectMessage(fileFragmentIndexMessage.sourceHandle, fileFragmentIndexMessage);                      
        }

    	/**
    	 * FileFragmentIndex retrieval request
    	 */
        else {            
        	fileFragmentIndexMessage.fileFragmentIndex = cdrmDataManager.getFileFragmentIndex(fileFragmentIndexMessage.fileId);
        	fileFragmentIndexMessage.isResponse = true;
        	
        	// Sends the address of the ADR to the source CDRM
        	virtualNode.sendDirectMessage(fileFragmentIndexMessage.sourceHandle, fileFragmentIndexMessage);           
        }
    }

}
