package br.usp.ime.oppstore.cdrm;

import org.apache.log4j.Logger;

import rice.p2p.commonapi.NodeHandle;
import br.usp.ime.oppstore.AdrAddress;
import br.usp.ime.oppstore.adrmanager.AdrManagerImpl;
import br.usp.ime.oppstore.message.StoreFragmentListMessage;
import br.usp.ime.oppstore.message.StoreFragmentMessage;
import br.usp.ime.virtualId.VirtualNode;
import br.usp.ime.virtualId.VirtualSpace;

public class FragmentStorageManager  {

    private AdrManagerImpl adrManager;
    private VirtualNode virtualNode;
    private Logger logger;
    
    public FragmentStorageManager(VirtualNode virtualNode) {
	
    	this.adrManager  = new AdrManagerImpl( virtualNode );
    	this.virtualNode = virtualNode;
    	this.logger = Logger.getLogger("adrManager." + virtualNode.getNode().getId().toString().substring(0, 9) + ">");    	
	}
    
    public AdrManagerImpl getAdrManager() { return adrManager; }
    
    void processStoreFragmentListMessage(StoreFragmentListMessage storeFragmentMessage) {
    	 
    	/**
    	 * TODO: We initially support only independent requests. Later, we will update the system to also support storage in the neighborhood. 
    	 */
    	
//        int nFragments = storeFragmentMessage.fragmentIdList.size();
//        Vector<NodeHandle> selectedNodes = new Vector<NodeHandle>(nFragments);
//                            
//        Id leftId  = null;
//        Id rightId = null;
//
//        long leftLongId  = idManipulator.convertIdToLong(leftId);
//        long rightLongId = idManipulator.convertIdToLong(rightId);
//
//        long longIdRange;
//        if (leftLongId <= rightLongId)
//        	longIdRange = rightLongId - leftLongId;
//        else
//        	longIdRange = (IdManipulator.MAX_ID - leftLongId) + rightLongId;
//
//        if (nFragments <= node.getLeafSet().getUniqueCount()) {
//        	Set<NodeHandle> nodeHandleSet = new HashSet<NodeHandle>();
//
//        	// Select random Ids from neighborMap
//        	while (nodeHandleSet.size() < nFragments) {
//        		Long randomlongId = (Math.abs(random.nextLong()) % longIdRange) + leftLongId;
//        		Id randomId = idManipulator.convertLongToId(randomlongId);
//        		NodeHandle randomNodeHandle = node.getAdaptiveIdTarget(randomId);
//        		nodeHandleSet.add(randomNodeHandle);                    
//        	}
//
//        	for (NodeHandle nh : nodeHandleSet)
//        		selectedNodes.add(nh);                                
//            
//        }  // OpStoreIdProtocol.ADAPTIVE      
//        
//        for (int fragmentIndex = 0;  fragmentIndex < nFragments; fragmentIndex++ ) {
//                        
//            Message storeMessage = new StoreFragmentMessage (
//                    storeFragmentMessage.fileId, 
//                    storeFragmentMessage.fragmentIdList.get(fragmentIndex), 
//                    storeFragmentMessage.fragmentSizeList.get(fragmentIndex), 
//                    storeFragmentMessage.sourceHandle, 
//                    storeFragmentMessage.messageNumber);
//            
//            // routes the message in the Network                
//            sendDirectMessage(selectedNodes.get(fragmentIndex), storeMessage);
//            
//        }
        
    }

    /**
     * @param message
     * @return true if an ADR from the current cluster was selected and false otherwise.
     */
    boolean processStoreFragmentMessage(StoreFragmentMessage message) {
                      
    	logger.debug("Received fragment store request for fragment " + message.fragmentId + ".");
    	    	
    	/**
    	 * Gets the address of a suitable ADR
    	 */
    	String adrAddress = adrManager.getAdrAddress(message.fragmentSize, message.fragmentId, message.timeoutMinutes);

    	/**
    	 * Returns the address of the ADR to the source CDRM.
    	 */ 
    	if (adrAddress != null || message.nStorageTrials == 0) {
    		message.setAdrAddress(adrAddress);
    		message.isResponse = true;
    		
    		logger.debug("Adr selected for fragment storage at " + adrAddress + ".");
    		
    		virtualNode.sendDirectMessage(message.sourceHandle, message);
    	}
    	
    	/**
    	 * If a suitable ADR is not found in the current cluster, forwards the request to another CDRM.
    	 */
    	else {
    		message.nStorageTrials--;
    		NodeHandle nextHandle = virtualNode.getNode().getLeafSet().get(1);

    		if (nextHandle != null) {
    			logger.debug("Adr for fragment storage could not be found! Redirecting storage request to node " + nextHandle.getId() + ".");
    			virtualNode.sendDirectMessage(nextHandle, message);
    		}
    		else
    			logger.debug("Adr for fragment storage could not be found!");
    	}

    	if (adrAddress != null) return true;
    	else return false;
    }

}
