package br.usp.ime.oppstore.simulation.simnode;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import org.omg.CORBA.ORB;
import org.omg.CORBA.Object;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAPackage.ServantNotActive;
import org.omg.PortableServer.POAPackage.WrongPolicy;

import rice.environment.Environment;
import rice.pastry.NodeIdFactory;

import br.usp.ime.oppstore.corba.AccessBroker;
import br.usp.ime.oppstore.corba.AccessBrokerHelper;
import br.usp.ime.oppstore.simulation.OppStoreSimulatorRemote;
import br.usp.ime.oppstore.simulation.broker.AccessBrokerSimulator;
import br.usp.ime.oppstore.simulation.broker.AccessBrokerSimulatorRemote;
import br.usp.ime.oppstore.statistics.BrokerStatisticsCollector;

public class MachineAccessBrokerManager {

    Environment env;
    
    Vector<AccessBrokerSimulator> accessBrokerSimulatorList;
    AccessBrokerSimulatorRemote[] accessBrokerRemoteList;
    
    OppStoreSimulatorRemote opStoreSimulatorRemote;
    BrokerStatisticsCollector brokerStatisticsCollector;    
    MachineCdrmAdrManager cdrmAdrManager;
    NodeIdFactory nodeIdFactory;
    
    POA poa;
    ORB orb;
    
    MachineAccessBrokerManager(Environment env, 
            OppStoreSimulatorRemote opStoreSimulatorRemote, 
            MachineCdrmAdrManager cdrmAdrManager,
            NodeIdFactory nodeIdFactory, ORB orb, POA poa) {
        
        this.env = env;
        this.opStoreSimulatorRemote = opStoreSimulatorRemote;
        this.brokerStatisticsCollector = new BrokerStatisticsCollector();
        this.cdrmAdrManager = cdrmAdrManager;
        this.nodeIdFactory = nodeIdFactory;
        this.poa = poa;
        this.orb = orb;
    }
    
    void createAccessBrokers(int numberOfAccessBrokers) {
        accessBrokerSimulatorList = new Vector<AccessBrokerSimulator>();        
        for (int i=0; i<numberOfAccessBrokers; i++) {
        	AccessBrokerSimulator simulator = 
        		new AccessBrokerSimulator( opStoreSimulatorRemote, cdrmAdrManager.getCdrmImpl(i), nodeIdFactory, brokerStatisticsCollector, i );
            accessBrokerSimulatorList.add( simulator );
            try {
				Object simulatorObj = this.poa.servant_to_reference( simulator );
				AccessBroker accessBroker = AccessBrokerHelper.narrow( simulatorObj );
				String accessBrokerIor = orb.object_to_string(accessBroker);
				//System.out.println("Starting AccessBroker with IOR:" + accessBrokerIor );
				simulator.setAccessBrokerIor( accessBrokerIor );
				
			} catch (Exception e) {
				e.printStackTrace();
			}
        }
        brokerStatisticsCollector.setNumberOfBrokers(accessBrokerSimulatorList.size());
    }
    
    void registerRemoteAccessBrokers() {
    
        try {
            accessBrokerRemoteList = new AccessBrokerSimulatorRemote[accessBrokerSimulatorList.size()];
            for (int i=0; i< accessBrokerRemoteList.length; i++) 
                accessBrokerRemoteList[i] = 
                    (AccessBrokerSimulatorRemote) UnicastRemoteObject.exportObject(
                            accessBrokerSimulatorList.get(i), 0);                        
            
        } catch (RemoteException e) {
            System.out.println("Could not register local ADRs. Aborting...");
            System.out.println(e.getMessage());
            env.destroy();
            System.exit(-1);
        }
    }
}
