package br.usp.ime.oppstore.simulation.simnode;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.rmi.RemoteException;
import java.util.Map.Entry;

import org.omg.CORBA.ORB;
import org.omg.PortableServer.POA;

import rice.environment.Environment;
import rice.pastry.Id;
import rice.pastry.NodeIdFactory;
import rice.pastry.PastryNode;
import br.usp.ime.oppstore.cdrm.CdrmApp;
import br.usp.ime.oppstore.cdrm.CdrmApp.OpStoreIdProtocol;
import br.usp.ime.oppstore.simulation.CdrmBootstrapInformation;
import br.usp.ime.oppstore.simulation.OppStoreSimulatorRemote;
import br.usp.ime.oppstore.simulation.adr.Adr;
import br.usp.ime.oppstore.simulation.adr.ClusterAdrSimulator;
import br.usp.ime.virtualId.VirtualSpace;

// Is likely to replace most of MachineCdrmAndBrokerLauncher
public class SimulationFileReader {

    private PrintStream cdrmStartLog;
    
    enum InputPosition {adr, cdrm, simulation, beginnning};
    
    MachineCdrmAdrManager cdrmAdrManager;
    MachineAccessBrokerManager accessBrokerManager;    
    OppStoreSimulatorRemote opStoreSimulator;
    
    CdrmBootstrapInformation bootInfo;
    NodeIdFactory nodeIdFactory;
    Environment env;    
    
    ORB orb;
    POA poa;
    
    int numberOfBrokers     = -1;
    int numberOfAdrs        = -1;
    int numberOfCdrms       = -1;
    int numberOfExperiments = -1;
        
    public SimulationFileReader(
            Environment env, NodeIdFactory nodeIdFactory, OppStoreSimulatorRemote opStoreSimulator, ORB orb, POA poa) {
        
    	this.orb = orb;
    	this.poa = poa;
        this.env = env;
        this.nodeIdFactory = nodeIdFactory;
        this.opStoreSimulator = opStoreSimulator;
        
        try {
            this.cdrmStartLog = 
                new PrintStream( new BufferedOutputStream( new FileOutputStream("cdrmStartup.log") ));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }

    }
        
    void createCdrmAdrFromFile (File inputFile, int localBindPort, CdrmBootstrapInformation bootInfo) {
        
        this.bootInfo = bootInfo;
        
        InputPosition inputPosition = InputPosition.beginnning;
        
        int cdrmNumber = -1;
        int adrNumber = -1;
        
        /**
         * Parse all file contents
         */
        BufferedReader bufferedReader = null;
        try {
            bufferedReader = new BufferedReader(new FileReader(inputFile));
            String line = null;
            while ( (line = bufferedReader.readLine()) != null ) {
                if (line.length() == 0 || line.charAt(0) == ' ' || line.charAt(0) == '#')
                    continue; 

                else if (line.startsWith("@ADR")) {
                    inputPosition = InputPosition.adr;
                    adrNumber++;
                }
                else if (line.startsWith("@CDRM")) {
                    inputPosition = InputPosition.cdrm;                                                           
                    if (cdrmNumber >= 0) {                        
                    	assert(adrNumber+1 == numberOfAdrs);                                            	
                        waitUntilCdrmIsReady( cdrmNumber );
                        if (cdrmNumber > 8 && CdrmApp.idProtocol == OpStoreIdProtocol.ADAPTIVE && cdrmNumber%1 == 0)
                            cdrmAdrManager.testVirtualIdRing();
                        /**
                         * Write statistics to file 
                         */
                        if (cdrmNumber > 3 && CdrmApp.idProtocol == OpStoreIdProtocol.ADAPTIVE)
                        	cdrmAdrManager.printVirtualIdStatistics();
                    }                                        
                    
                    /**
                     * Kills a CDRM
                     */
                    if (cdrmNumber%10 == -8) {                    	
                    	cdrmAdrManager.destroyLocalCdrm(cdrmNumber/2);                        
                        try { Thread.sleep(60 * 1000); }
                        catch (InterruptedException e) {}
                    }
                    /**
                     * Changes an CDRM capacity
                     */
                    if (cdrmNumber%10 == -5) {
                    	int killCdrmNumber = cdrmNumber/2 + cdrmNumber%2;
                    	System.out.println("Updating ADR values at node " + cdrmAdrManager.getCdrmApp(killCdrmNumber).getVirtualNode().getNode().getId() + ".");
                    	ClusterAdrSimulator clusterAdrSimulator = cdrmAdrManager.getClusterAdrSimulator(killCdrmNumber); 
                    	for (Entry< Integer, Adr> entry: clusterAdrSimulator.getAdrList() ) {
                    		Adr adr = entry.getValue();
                    		clusterAdrSimulator.updateAdrParameter(entry.getKey(), (int)(adr.getFreeStorageSpace()/2.0), adr.getMeanUptime(), adr.getMeanIdleness());
                    	}
                        try { Thread.sleep( 2 * 1000 ); }
                        catch (InterruptedException e) {}
                    }
                    cdrmNumber++;
                    adrNumber = -1;
                } 
                else if (line.startsWith("@SIMULATION")) {
                    inputPosition = InputPosition.simulation;
                }

                else if (inputPosition == InputPosition.adr) {  
                    processAdrInfo(line, cdrmNumber, adrNumber);
                }
                else if (inputPosition == InputPosition.cdrm) {
                    processCdrmInfo(line, cdrmNumber);
                }
                else if (inputPosition == InputPosition.simulation) {
                    processSimulationInfo(line, localBindPort, bootInfo);
                }
                
            }            
        } catch (IOException e) {
            e.printStackTrace();
        }              
        
        assert(cdrmNumber+1 == numberOfCdrms);
                        
        /**
         * If using the virtual ids:
         * - Waits until all nodes are ready
         * - Tests the virtual id ring for consistency
         */
        this.waitUntilCdrmIsReady(cdrmNumber);
        if (CdrmApp.idProtocol == OpStoreIdProtocol.ADAPTIVE) {
            cdrmAdrManager.testVirtualIdRing();
            cdrmAdrManager.printVirtualIdStatistics();
        }
        	        
        System.out.println("Finished setting up clusters and adrs.");
        
        /**
         * Creates the access Brokers to perform the simulation 
         */
        System.out.println("Creating the access brokers.");
        accessBrokerManager = new MachineAccessBrokerManager(env, opStoreSimulator, cdrmAdrManager, nodeIdFactory, orb, poa);
        accessBrokerManager.createAccessBrokers(numberOfBrokers);
        accessBrokerManager.registerRemoteAccessBrokers();        

        try { opStoreSimulator.setRemoteMachineStarted(accessBrokerManager.accessBrokerRemoteList); } 
        catch (RemoteException e) { e.printStackTrace(); }
                
    }

    /**
     * Waits until the previous CDRM is ready.
     */
	private void waitUntilCdrmIsReady(int cdrmNumber) {
		if (CdrmApp.idProtocol == CdrmApp.OpStoreIdProtocol.ADAPTIVE && cdrmAdrManager.getNumberOfAdrsInCdrm(cdrmNumber) > 0) {
			VirtualSpace virtualSpace = cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getVirtualSpace(1);
			while ( virtualSpace == null || virtualSpace.getVirtualLeafSet().isUpdated() == false ) {
				try { Thread.sleep(100); }
				catch (InterruptedException e) {}
				virtualSpace = cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getVirtualSpace(1);
			}
            try { Thread.sleep( (cdrmNumber/50 + 1) * 2000 ); }
            catch (InterruptedException e) {}            
		}
		else if (CdrmApp.idProtocol == CdrmApp.OpStoreIdProtocol.PASTRY) {
			while ( cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode().isReady() == false ) {
				try { Thread.sleep(100); }
				catch (InterruptedException e) {}
			}                        	
		}
	}

    private void processSimulationInfo(String line, int localBindPort, CdrmBootstrapInformation bootInfo) {
        
        String[] simulationStringInfo = line.split("[= ]");        
        assert(simulationStringInfo.length == 6);
                        
        // create CDRMs        
        assert (simulationStringInfo[0].compareTo("numberOfCdrms") == 0);        
        numberOfCdrms = Integer.parseInt(simulationStringInfo[1]);                
        this.cdrmAdrManager = new MachineCdrmAdrManager(env, localBindPort, nodeIdFactory, orb); 
        //cdrmAdrManager.createLocalCdrms(bootInfo.bootstrapNodeAddress, numberOfCdrms, 0);
                
        // create Brokers
        assert (simulationStringInfo[2].compareTo("numberOfBrokers") == 0);        
        numberOfBrokers = Integer.parseInt(simulationStringInfo[3]);        

        assert (simulationStringInfo[4].compareTo("numberOfExperiments") == 0);        
        numberOfExperiments = Integer.parseInt(simulationStringInfo[5]);        
    }
    
    /**
     * Creates a CDRM and the ClusterAdrSimulator from the information contained in the file.
     * 
     * @param line
     * @param cdrmNumber
     */
    private void processCdrmInfo(String line, int cdrmNumber) {

        String[] simulationStringInfo = line.split("[= ]");
        assert(simulationStringInfo.length == 4);
       
        // create clusterAdrSimulator
        assert (simulationStringInfo[0].compareTo("numberOfAdrs") == 0);        
        numberOfAdrs = Integer.parseInt(simulationStringInfo[1]);        
        
        if (cdrmNumber > 0) {
            PastryNode cdrmNode = cdrmAdrManager.getCdrmApp(cdrmNumber-1).getVirtualNode().getNode();
            while (!cdrmNode.isReady())
                try { Thread.sleep(100); } catch (InterruptedException e) {}            
        }
        
        Id nodeId = null;       
        // nodeId = Id.build (new byte[]{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,-127});

        cdrmAdrManager.createLocalCdrm(cdrmNumber, nodeId, bootInfo.bootstrapNodeAddress, 0);
        cdrmAdrManager.createClusterAdrSimulator(cdrmNumber, cdrmAdrManager.getCdrmApp(cdrmNumber), numberOfAdrs);
        
        System.out.println(cdrmNumber + ": " + cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode() + " [" + 
                cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode().getId().toByteArray()[19] + "," +
                cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode().getId().toByteArray()[18] + "] ");

        cdrmStartLog.println(cdrmNumber + ": " + cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode() + " [" + 
                cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode().getId().toByteArray()[19] + "," +
                cdrmAdrManager.getCdrmApp(cdrmNumber).getVirtualNode().getNode().getId().toByteArray()[18] + "] ");
        cdrmStartLog.flush();

    }    

    /**
     * Reads the ADR info from input file.
     * freeSpace=3523 timeOffset=0 meanDayUptime=0.1 meanNightUptime=0.1 meanDayIdleness=0.1 meanNightIdleness=0.1
     */ 
    private void processAdrInfo(String line, int cdrmNumber, int adrNumber) {

        String[] simulationStringInfo = line.split("[= ]");
        assert(simulationStringInfo.length == 12);
        
        assert (simulationStringInfo[0].compareTo("freeSpace") == 0);                   
        int freeSpace = Integer.parseInt(simulationStringInfo[1]);
        assert (simulationStringInfo[2].compareTo("timeOffset") == 0);                   
        int timeOffset = Integer.parseInt(simulationStringInfo[3]);            
        
        assert (simulationStringInfo[4].compareTo("meanDayUptime") == 0);        
        double meanDayUptime = Double.parseDouble(simulationStringInfo[5]);
        assert (simulationStringInfo[6].compareTo("meanNightUptime") == 0);        
        double meanNightUptime = Double.parseDouble(simulationStringInfo[7]);
        
        assert (simulationStringInfo[8].compareTo("meanDayIdleness") == 0);        
        double meanDayIdleness = Double.parseDouble(simulationStringInfo[9]);
        assert (simulationStringInfo[10].compareTo("meanNightIdleness") == 0);        
        double meanNightIdleness = Double.parseDouble(simulationStringInfo[11]);
        
        //System.out.println("AdrNumber = " + adrNumber + " of " + numberOfAdrs + " " + cdrmNumber);
        
        //if (cdrmAdrManager.getNumberOfCdrms() == cdrmNumber + 1) {
        
        		//System.out.println( "numberOfCdrmsr:" + cdrmAdrManager.getNumberOfCdrms() + " cdrmNumber:" + cdrmNumber );
        		ClusterAdrSimulator clusterAdrs = cdrmAdrManager.getClusterAdrSimulator(cdrmNumber);
        		clusterAdrs.setAdrParameters(adrNumber, freeSpace, numberOfExperiments, timeOffset,
        				meanDayUptime, meanNightUptime, meanDayIdleness, meanNightIdleness);
        	
        		if (adrNumber == numberOfAdrs - 1)
        			clusterAdrs.registerAdrs();
        //}
    }    
}
